//
// GameX - ImageX Class Code 
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#include <math.h>
#include <string.h>

#define GAMEX_WINDOWS
#include "gamex.hpp"

ImageX::ImageX (void)
{
	m_xres = 0; m_yres = 0;
	m_options = IMG_UNDEF;
	m_data = NULL;
	m_alpha = NULL;
	m_usage = DRAW_UNDEF;
	status = IMG_NOTREADY;
}

ImageX::ImageX (int xr, int yr, ImageOps ops)
{
	Size (xr, yr, ops);
}

ImageX::~ImageX (void)
{
	if (m_data!=NULL) delete [] m_data;
	if (m_alpha!=NULL) delete [] m_alpha;
	m_xres = 0; m_yres = 0;
	m_options = IMG_UNDEF;
	m_data = NULL;
	m_alpha = NULL;	
}

inline int ImageX::GetBytesPerPixel (ImageOps ops)
{												// Pixel Format
	if (ops & IMG_MERGEALPHA) {					// I R G B A
		if (ops & IMG_INDEXED) return 2;		// 8-0-0-0-8
		if (ops & IMG_LOWCOLOR) return 2;		// 0-4-4-4-4
		if (ops & IMG_HIGHCOLOR) return 3;		// 0-6-5-5-8
		if (ops & IMG_TRUECOLOR) return 4;		// 0-8-8-8-8
	} else {
		if (ops & IMG_INDEXED) return 1;		// 8-0-0-0-0
		if (ops & IMG_LOWCOLOR) return IMG_UNDEF;// undefined
		if (ops & IMG_HIGHCOLOR) return 2;		// 0-6-5-5-0
		if (ops & IMG_TRUECOLOR) return 3;		// 0-8-8-8-0
	}
	return IMG_UNDEF;
}

inline int ImageX::GetFormat (ImageOps ops)
{
	if (ops & IMG_INDEXED) return IMG_INDEXED;
	if (ops & IMG_LOWCOLOR) return IMG_LOWCOLOR;
	if (ops & IMG_HIGHCOLOR) return IMG_HIGHCOLOR;
	if (ops & IMG_TRUECOLOR) return IMG_TRUECOLOR;
	return IMG_UNDEF;
}

inline int ImageX::GetAlpha (ImageOps ops)
{	
	if (ops & IMG_ALPHA) return IMG_ALPHA;
	return IMG_UNDEF;
}

inline void ImageX::TurnOff (ImageOps opt)
{	
	m_options ^= opt;
}

inline void ImageX::TurnOn (ImageOps opt)
{	
	m_options |= opt;
}

void ImageX::Size (int xr, int yr, ImageOps ops)
{
	int bpp;	
	
	bpp = GetBytesPerPixel (ops);
	
	if (bpp!=-1) {
		if (m_data!=NULL) delete [] m_data;	
		m_data = new XBYTE[xr * yr * bpp];
	}
	if (ops && IMG_ALPHA) {
		if (m_alpha!=NULL) delete [] m_alpha;
		m_alpha = new XBYTE[xr * yr];
	}	
	m_xres = xr; m_yres = yr;
	m_options = ops;
}

void ImageX::Create (int xr, int yr, int what)
{	
	switch (what) {
	case DRAW_IMAGE: Size (xr, yr, IMG_HIGHCOLOR); break;
	case DRAW_MASKED: Size (xr, yr, IMG_HIGHCOLOR | IMG_ALPHA); break;
	case DRAW_ALPHA: case DRAW_ADDED: Size (xr, yr, IMG_HIGHCOLOR | IMG_ALPHA); break;	
	}
	Fill (0,0,0);
	m_usage = what;
	status = IMG_OK;
}

void ImageX::Fill (int r, int g, int b)
{
	XBYTE2 *pData;
			
	pData = (XBYTE2*) m_data;
	for (int n=0; n < m_xres*m_yres; n++) {
		*pData++ = (r << win_shiftr) + (g << win_shiftg) + (b << win_shiftb);
	}
}

void ImageX::UseWith (int what)
{
	m_usage = what;
	ConvertFormat (IMG_HIGHCOLOR);
	switch (what) {
	case DRAW_IMAGE:					break;
	case DRAW_MASKED:					ConvertAlphaToMask(); break;
	case DRAW_ALPHA: case DRAW_ADDED:	ConvertAlphaToBlend(); break;
	}
	status = IMG_OK;
}

void ImageX::SetPixel (int x, int y, int r, int g, int b)
{
	if (x >=0 && y>=0 && x < m_xres && y < m_yres) {
		XBYTE2 *pData = (XBYTE2*) m_data;	
		pData += (y*m_xres)+x;
		*pData = (r << win_shiftr) + (g << win_shiftg) + (b << win_shiftb);
	}
}

void ImageX::GetPixel (int x, int y, int &r, int &g, int &b)
{
	if (x >=0 && y>=0 && x < m_xres && y < m_yres) {
		XBYTE2 *pData = (XBYTE2*) m_data;
		pData += (y*m_xres)+x;
		r = (*pData & win_maskr) >> win_shiftr;
		g = (*pData & win_maskg) >> win_shiftg;
		b = (*pData & win_maskb) >> win_shiftb;	
	} else {
		r = 0; g = 0; b = 0;
	}
}

void ImageX::ConvertFormat (ImageOps ops)
{	
	int n;
	int v;
	XBYTE * newdata;
	
	switch (GetFormat(m_options)) {			// Source Format
	case IMG_TRUECOLOR: {					//   True Color (24 bit)
		switch (GetFormat(ops))	{			// Target Format
		case IMG_HIGHCOLOR: {				//   High Color (16 bit)
			XBYTE *pDataOld;
			XBYTE2 *pDataNew;
			newdata = new XBYTE[m_xres * m_yres * GetBytesPerPixel(ops)];			
			pDataOld = (XBYTE *) m_data; 
			pDataNew = (XBYTE2 *) newdata;
			for (n=0; n<m_xres*m_yres; n++) {
				v = (*pDataOld++ << win_maxr) >> 8; *pDataNew = v << win_shiftr;
				v = (*pDataOld++ << win_maxg) >> 8;	*pDataNew |= v << win_shiftg;
				v = (*pDataOld++ << win_maxb) >> 8; *pDataNew++ |= v << win_shiftb;
			}
			delete [] m_data;
			m_data = newdata;			
		} break;
		}
	} break;
	}
	TurnOff (GetFormat(m_options));
	TurnOn (GetFormat(ops));
}

void ImageX::ConvertAlphaToMask (void)
{
	// NOTE: IF NO ALPHA CHANNEL IS LOADED, any pixels with value 0 
	// in the primary data with be treated as transparent. DrawMasked
	// will draw all 0 value pixels transparent. This is the desired behavior.
	// If an alpha channel >is< present, 0 value pixels will be mapped to value 1,
	// and the alpha channel will be used to determine which pixels to make transparent
	// by setting them to 0.
	if (GetAlpha(m_options)==IMG_ALPHA) {
		switch (GetFormat(m_options)) {
		case IMG_HIGHCOLOR: {
			XBYTE2 *pData;
			XBYTE *pAlphaData;
			pData = (XBYTE2*) m_data;
			pAlphaData = m_alpha;
			for (int n=0; n<m_xres*m_yres; n++) {
				if (*pData==0) *pData = 1;
				if (*pAlphaData==0) *pData = 0;
				*pData++;
				*pAlphaData++;
			}
		} break;
		}
		delete [] m_alpha;
		m_alpha = NULL;
		TurnOff (IMG_ALPHA);
		TurnOn (IMG_MASKED);
	}
}

void ImageX::ConvertAlphaToBlend (void)
{
	if (GetAlpha(m_options)==IMG_ALPHA) {		
		switch (GetFormat(m_options)) {
		case IMG_TRUECOLOR: {			
			XBYTE *pData;
			XBYTE *pAlphaData;
			pData = (XBYTE*) m_data;
			pAlphaData = m_alpha;
			for (int n=0; n<m_xres*m_yres; n++) {
				*pData++ = (*pData * *pAlphaData)>>8;
				*pData++ = (*pData * *pAlphaData)>>8;
				*pData++ = (*pData * *pAlphaData)>>8;
				*pAlphaData++ = 255 - *pAlphaData;
			}
		} break;
		case IMG_HIGHCOLOR: {
			XBYTE2 r, g, b;
			XBYTE2 *pData;
			XBYTE *pAlphaData;			
			pData = (XBYTE2*) m_data;
			pAlphaData = m_alpha;			
			for (int n=0; n<m_xres*m_yres; n++) {				
				r = (((*pData & win_maskr) >> win_shiftr) * *pAlphaData) >> 8;
				g = (((*pData & win_maskg) >> win_shiftg) * *pAlphaData) >> 8;
				b = (((*pData & win_maskb) >> win_shiftb) * *pAlphaData) >> 8;				
				*pData++ = (r << win_shiftr) + (g << win_shiftg) + (b << win_shiftb);
				*pAlphaData = 255 - *pAlphaData;
				pAlphaData++;
			}
		} break;
		}		
	} else {										// NO ALPHA CHANNEL LOADED
		if (m_alpha!=NULL) delete [] m_alpha;	
		m_alpha = new XBYTE[m_xres * m_yres];		// Create an alpha channel	
		TurnOn (IMG_ALPHA);		
		switch (GetFormat(m_options)) {
		case IMG_TRUECOLOR: {			
			XBYTE *pData;
			XBYTE *pAlphaData;
			pData = (XBYTE*) m_data;
			pAlphaData = m_alpha;
			for (int n=0; n<m_xres*m_yres; n++) {	// Alpha channel is gray of primary data
				*pAlphaData = (*pData + *(pData+1) + *(pData+2))/3;
				*pData++ = (*pData * *pAlphaData)>>8;
				*pData++ = (*pData * *pAlphaData)>>8;
				*pData++ = (*pData * *pAlphaData)>>8;
				*pAlphaData++ = 255 - *pAlphaData;
			}
		} break;
		case IMG_HIGHCOLOR: {
			XBYTE2 r, g, b;
			XBYTE2 *pData;
			XBYTE *pAlphaData;			
			pData = (XBYTE2*) m_data;
			pAlphaData = m_alpha;			
			for (int n=0; n<m_xres*m_yres; n++) {				
				r = (*pData & win_maskr) >> win_shiftr;
				g = (*pData & win_maskg) >> win_shiftg;
				b = (*pData & win_maskb) >> win_shiftb;
				*pAlphaData = ((r + g + b)*4)/3;
				r = (r * *pAlphaData) >> 8;
				g = (g * *pAlphaData) >> 8;
				b = (b * *pAlphaData) >> 8;
				*pData++ = (r << win_shiftr) + (g << win_shiftg) + (b << win_shiftb);
				*pAlphaData = 255 - *pAlphaData;
				pAlphaData++;
			}
		} break;
		}
	}
}

void ImageX::Rotate (ImageX *dest, float ang, float scale)
{	
	dest->SetUsage (m_usage);
	RotateImage (dest, ang, scale);
	RotateAlpha (dest, ang, scale);
}
		
void ImageX::RotateImage (ImageX *dest, float ang, float scale)
{
	// NOTE: ASSUMES BOTH IMAGES ARE IMG_HIGHCOLOR

	XBYTE2 *src_pix;
	XBYTE2 *dest_pix;
	XBYTE2 *dest_start;
	float delt_x, delt_y;
	float start_x, start_y;
	float off_x, off_y;
	float pix_x, pix_y, h;
	int y, cnt;

	if (dest!=NULL && m_data!=NULL) {
	
	dest->SetUsage(m_usage);
	h = (float) sqrt(dest->m_xres * dest->m_xres + dest->m_yres * dest->m_yres);
	start_x = (m_xres/2.0f) + (float) cos((135-ang)*DEGtoRAD)*(h/2.0f)/scale;
	start_y = (m_yres/2.0f) - (float) sin((135-ang)*DEGtoRAD)*(h/2.0f)*(m_xres/m_yres)/scale;
	delt_x = (float) cos(ang*DEGtoRAD)/scale;
	delt_y = (float) sin(ang*DEGtoRAD)*(m_xres/m_yres)/scale;
	off_x = delt_x * dest->m_xres;
	off_y = delt_y * dest->m_xres;	
	dest_start = (XBYTE2*) dest->m_data;	
	for (y=0; y<dest->m_yres; y++) {	
		cnt = dest->m_xres;						// Start with full pixel set in line
		pix_x = start_x + off_x;				// Goto end of line and scan backwards
		pix_y = start_y + off_y;
		dest_pix = dest_start + cnt-1;
		while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
			pix_x -= delt_x;					// Scan until end of span found
			pix_y -= delt_y;
			cnt--;
			*dest_pix-- = 0;
		}	
		dest_pix = dest_start;
		pix_x = start_x;						// Goto beginning of line and scan forward
		pix_y = start_y;
		while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
			pix_x += delt_x;					// Scan until start of span found
			pix_y += delt_y;
			cnt--;
			*dest_pix++ = 0;
		}
		while (cnt>0) {
			src_pix = ((XBYTE2*) m_data) + (int) pix_y*m_xres + (int) pix_x;
			*dest_pix++ = *src_pix;			
			pix_x += delt_x; pix_y += delt_y;
			cnt--;
		}
		dest_start += dest->m_xres;
		start_x -= delt_y;
		start_y += delt_x;
	}

	}
}

void ImageX::RotateAlpha (ImageX *dest, float ang, float scale)
{
	XBYTE *src_pix;
	XBYTE *dest_pix;
	XBYTE *dest_start;
	float delt_x, delt_y;
	float start_x, start_y;
	float off_x, off_y;
	float pix_x, pix_y, h;
	int y, cnt;

	if (m_alpha!=NULL && dest->m_alpha!=NULL) {

		dest->SetUsage(m_usage);
		h = (float) sqrt(dest->m_xres * dest->m_xres + dest->m_yres * dest->m_yres);
		start_x = (m_xres/2.0f) + (float) cos((135-ang)*DEGtoRAD)*(h/2.0f)/scale;
		start_y = (m_yres/2.0f) - (float) sin((135-ang)*DEGtoRAD)*(h/2.0f)/scale;
		delt_x = (float) cos(ang*DEGtoRAD)/scale;
		delt_y = (float) sin(ang*DEGtoRAD)/scale;
		off_x = delt_x * dest->m_xres;
		off_y = delt_y * dest->m_xres;	
		dest_start = (XBYTE*) dest->m_alpha;
		for (y=0; y<dest->m_yres; y++) {	
			cnt = dest->m_xres;						// Start with full pixel set in line
			pix_x = start_x + off_x;				// Goto end of line and scan backwards
			pix_y = start_y + off_y;
			dest_pix = dest_start + cnt-1;
			while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
				pix_x -= delt_x;					// Scan until end of span found
				pix_y -= delt_y;
				cnt--;
				*dest_pix-- = 255;
			}	
			dest_pix = dest_start;
			pix_x = start_x;						// Goto beginning of line and scan forward
			pix_y = start_y;
			while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
				pix_x += delt_x;					// Scan until start of span found
				pix_y += delt_y;
				cnt--;
				*dest_pix++ = 255;
			}
			while (cnt>0) {
				src_pix = ((XBYTE*) m_alpha) + (int) pix_y*m_xres + (int) pix_x;
				*dest_pix++ = *src_pix;			
				pix_x += delt_x; pix_y += delt_y;
				cnt--;
			}
			dest_start += dest->m_xres;
			start_x -= delt_y;
			start_y += delt_x;
		}
	}
}


ImageExt::ImageExt ()
{
	m_pImg = NULL;
}

ImageExt::ImageExt (ImageX* pImg)
{
	m_pImg = pImg;
}

ImageExt::~ImageExt ()
{
	m_pImg = NULL;
}

void ImageExt::Load (char *filename, ImageX *img, int what)
{
	LoadTiff (filename, img);
	img->UseWith (what);	
}

#ifdef IMAGEX_SUPPORT_RAW
	#include "gamex-image-raw.hpp"
	#include "gamex-image-raw.cpp"
#endif

#ifdef IMAGEX_SUPPORT_TIFF
	#include "gamex-image-tiff.hpp"
	#include "gamex-image-tiff.cpp"
#endif