//
// GameX - WindowsDX Class Code (including WinMain and WinProc)
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#include "gamex-sound.hpp"
#include "gamex-win-dx.hpp"

// WindowsDX Constructor
WindowsDX::WindowsDX ()
{	
	win_request_xsize = -1;
	win_request_ysize = -1;
	win_status = GAMEX_NO_INIT;
	table_alpha = NULL;
	num_sounds = 0;
	max_sounds = 0;
	sound_list = NULL;
	ds_sound_list = NULL;
	sound_support = false;
}

WindowsDX::~WindowsDX (void)
{
	if (table_alpha!=NULL) delete [] table_alpha;
}

// WinMain and WindowFunc functions supplied by WindowsDX
LRESULT CALLBACK WindowFunc (HWND hWnd,	UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) {
	case WM_ACTIVATEAPP: GameX.SetActivation ((BOOL) wParam); break;
	case WM_MOVE: GameX.SetPosition (LOWORD(lParam), HIWORD(lParam)); break;
	case WM_CHAR: {		
		GameX.SetKey ((TCHAR) wParam); 
	} break;
	case WM_KEYDOWN: GameX.SetKeyDown ((TCHAR) wParam); break;		
	case WM_KEYUP: GameX.SetKeyUp ((TCHAR) wParam); break;
	case WM_CREATE:	break;
	case WM_SETFOCUS: {
		if (GameX.GetStatus()==GAMEX_READY) {
			GameX.SetActivation (true);
			GameX.ViewUpdate(); 		
		}
	} break;
	case WM_KILLFOCUS: GameX.SetActivation(false); break;
	case WM_PAINT: break;
	case WM_SIZE: break;
	case WM_CLOSE: DestroyWindow (hWnd); break;
	case WM_DESTROY: PostQuitMessage (0); break;
	case WM_MOUSEMOVE:		GameX.SetMouseMove (LOWORD (lParam), HIWORD (lParam));	break;
	case WM_LBUTTONDOWN:	GameX.SetMousePress (MK_LBUTTON);	break;
	case WM_MBUTTONDOWN:	GameX.SetMousePress (MK_MBUTTON);	break;
	case WM_RBUTTONDOWN:	GameX.SetMousePress (MK_RBUTTON);	break;
	case WM_RBUTTONUP: case WM_MBUTTONUP: case WM_LBUTTONUP:	GameX.SetMouseRelease ();			break;
	}
	
	return DefWindowProc (hWnd, message, wParam, lParam);		
}

int WINAPI WinMain (HINSTANCE hThisInst, HINSTANCE hPrevInst, LPSTR lpszArgs, int nWinMode)
{
	MSG msg;
	
	GameX.WindowInitInstance (hThisInst, nWinMode);	

	initialize ();

	if (GameX.GetStatus()!=GAMEX_READY) GameX.ReportProblem ();	
	
	for (;;) {
		if (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else {
			main();
		}
	}
	GameX.CleanUp ();
	return msg.wParam;
}

// WindowsDX Member Functions

void WindowsDX::Quit (void)
{
	PostQuitMessage (0);
}

bool WindowsDX::Initialize (int nWinFull, int xres, int yres)
{
	win_request_full = nWinFull;
	win_request_xsize = xres;
	win_request_ysize = yres;	
	
	debug.Open ("debug.txt");
	debug.Output ("WindowsDX: Initialization started.\n", "");	

	if (GameX.InitWindows ()==false)	return false;
	if (GameX.InitDirectX ()==false)	return false;
	if (GameX.InitSoundX ()==false)		return false;

	debug.Output ("WindowsDX: Initialization finished OK.\n", "");
	return true;
}

bool WindowsDX::InitWindows (void)
{	
	if (WindowInitClass()==false)		return false;
	if (WindowInit()==false)			return false;
	if (WindowDraw()==false)			return false;
	return true;
}

bool WindowsDX::InitDirectX (void)
{	
	if (DirectDrawInit ()==false)		return false;
	if (DirectDrawSurfInit ()==false)	return false;
	win_keys = 0;	
	return true;
}

bool WindowsDX::InitSoundX (void)
{
	if (DirectSoundInit ()==false)		return false;	
	return true;
}

void WindowsDX::WindowInitInstance (HINSTANCE hThisInst, int nWinMode)
{
	win_hinst = hThisInst;
	win_hmode = nWinMode;
}
		
bool WindowsDX::WindowInitClass (void)
{
	WNDCLASS wcl;

  	wcl.hCursor			= LoadCursor(NULL,IDC_ARROW);
	wcl.hIcon			= NULL;
    wcl.lpszMenuName	= "Menu";
    wcl.lpszClassName	= "GameX";
	wcl.hInstance		= win_hinst;
	wcl.style			= CS_VREDRAW | CS_HREDRAW;
    wcl.lpfnWndProc		= (WNDPROC) WindowFunc;
	wcl.cbClsExtra		= 0;
    wcl.cbWndExtra		= 0;
    wcl.hbrBackground	= (HBRUSH) GetStockObject(BLACK_BRUSH);
    
    if (!RegisterClass(&wcl)) {
		win_status = GAMEX_INITCLASS_FAILED;
		return false;
	}
	win_status = GAMEX_INITCLASS_OK;
	return true;
}

bool WindowsDX::WindowInit (void)
{
	DWORD style;
	HDC hdc;

	// Initialize View Format (Bits-Per-Pixel)
	hdc = GetDC (win_hwnd);
	win_screenbpp = GetDeviceCaps (hdc, BITSPIXEL);
	ReleaseDC (win_hwnd, hdc);

	if (win_screenbpp!=16) {
		win_status = GAMEX_BPP_FAILED;
		return false;
	}
	
	// Initialize View Size & FullScreen Mode
	win_screenx = GetSystemMetrics(SM_CXSCREEN);
	win_screeny = GetSystemMetrics(SM_CYSCREEN);
	win_fullscreen = win_request_full;
	
	// Initialize View Width & Height
	if (win_fullscreen==TRUE)	{
		style = WS_POPUP;	
		win_width = win_request_xsize;
        win_height = win_request_ysize;
	} else {
		if (win_request_xsize > win_screenx || win_request_ysize > win_screeny) {		
			win_status = GAMEX_REQSIZE_FAILED;
			return false;
		}
		style = WS_OVERLAPPED | WS_SYSMENU;
		win_width = win_request_xsize;		
		win_height = win_request_ysize;		
		if (win_width<=0 || win_width>win_screenx) win_width = win_screenx;
		if (win_height<=0 || win_height>win_screeny) win_height = win_screeny;			
	}

	// Initialize Drawing Viewport
	win_viewx1 = 0;
	win_viewy1 = 0;
	win_viewx2 = win_width-1;
	win_viewy2 = win_height-1;
		
	// Create window adjusting for desired client area size
	// NOTE: For some reason (windows crap??) the AdjustWindowRectEx function
	// is not working.
	RECT	rect;
	int		sx, sy;
	HWND	temp_hwnd;

	// Create temporary window to determine extended window size
	temp_hwnd = CreateWindowEx(WS_EX_APPWINDOW, "GameX", "", style, 0, 0, win_width, win_height, 0, 0, win_hinst, 0);
	if (temp_hwnd==NULL) {
		win_status = GAMEX_INITWINTEMP_FAILED;
		return false;		
	}
	win_status = GAMEX_INITWINTEMP_OK;
	GetClientRect (temp_hwnd, &rect);
	
	sx = (win_width<<1) - rect.right;
	sy = (win_height<<1) - rect.bottom;
	if (win_fullscreen==TRUE) {
		rect.left = 0; rect.top = 0;
	} else {
		rect.left = (win_screenx>>1) - (sx>>1);
		rect.top = (win_screeny>>1) - (sy>>1);
	}	
	rect.right = rect.left + sx-1;		
	rect.bottom = rect.top + sy-1;

	// Create actual window (with correct size) which will be used

	win_hwnd = CreateWindowEx(
					WS_EX_APPWINDOW, 
					"GameX", 
					"", 
					style, 
					rect.left, rect.top, 
					rect.right-rect.left, rect.bottom-rect.top,
					0, 
					0, 
					win_hinst, 
					0);
	if (win_hwnd==NULL) {
		win_status = GAMEX_INITWINREAL_FAILED;		
		return false;
	}

	// Double check to make sure we have correct client rect size	
	GetClientRect (win_hwnd, &rect);	
	
	win_status = GAMEX_INITWINREAL_OK;
	return true;
}

bool WindowsDX::WindowDraw (void)
{
	ShowWindow (win_hwnd, win_hmode);
	if (UpdateWindow (win_hwnd)==false) {
		win_status = GAMEX_SHOWWIN_FAILED;		
		return false;
	}	
	win_status = GAMEX_SHOWWIN_OK;
	return true;
}

bool WindowsDX::DirectDrawInit ()
{
	HRESULT status;
	int style;

	// Direct Draw 3.0	
	status = DirectDrawCreate (NULL, &dd_main, NULL);
	if (status!=DD_OK) {
		win_status = GAMEX_DDSTART_FAILED;
		return false;
	}
	win_status = GAMEX_DDSTART_OK;

	if (win_fullscreen==TRUE) {
		style = DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE;
	} else {
		style = DDSCL_NORMAL;
	}

	status = dd_main->SetCooperativeLevel (win_hwnd, style);
	if (status!=DD_OK) {
		win_status = GAMEX_DDSETLEVEL_FAILED;
		return false;
	}
	win_status = GAMEX_DDSETLEVEL_OK;
	return true;
}

bool WindowsDX::DirectSoundInit (void)
{
	if (DS_OK==DirectSoundCreate(NULL, &ds_main, NULL)) {
		if (DS_OK!=IDirectSound_SetCooperativeLevel(ds_main, win_hwnd, DSSCL_NORMAL)) {
			IDirectSound_Release(ds_main);
			win_status = GAMEX_SOUNDCOOP_FAILED;
			sound_support = false;
			return false;
		}
	} else {		
		win_status = GAMEX_SOUNDINIT_FAILED;
		sound_support = false;
        return false;
    }
	sound_support = true;
	return true;
}

int WindowsDX::GetShift (int mask)
{
	int count;
	for (count=0; (mask & 1) == 0; count++)
		mask >>= 1;
	return count;
}

int WindowsDX::GetMax (int mask)
{
	int count, max;	
	for (count=0; (mask & 1) == 0; count++)
		mask >>= 1;
	for (max=0; mask > 0; max++)
		mask >>= 1;
	return max;
}


bool WindowsDX::DirectDrawSurfInit ()
{	
	// DirectX 3.1
	HRESULT status;
	DDSURFACEDESC ddsd;	
	POINT pt;

	if (dd_back_buf)	dd_back_buf->Release (),	dd_back_buf = NULL;
	if (dd_front_buf)	dd_front_buf->Release (),	dd_front_buf = NULL;	

	ZeroMemory (&ddsd, sizeof (ddsd));
	
	if (win_fullscreen==TRUE) {
		// Set display mode
	 	status = dd_main->SetDisplayMode (win_width, win_height, win_screenbpp);
		if (status!=DD_OK) {
			win_status = GAMEX_SETMODE_FAILED;
			return false;
		}
		win_status = GAMEX_SETMODE_OK;
		// Set front surface description
		ddsd.dwSize = sizeof (ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.dwBackBufferCount = 1;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
		// Try to create primary surface 
		status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL);
		// No good, try using system memory
		if (status!=DD_OK) {
			switch (status) {
			case DDERR_INVALIDOBJECT:			debug.Output ("<WinDX::CreateSurface> Invalid Object\n","");break;
			case DDERR_INVALIDPARAMS:			debug.Output ("<WinDX::CreateSurface> Invalid Params\n","");break;
			case DDERR_OUTOFVIDEOMEMORY:		debug.Output ("<WinDX::CreateSurface> Out Of Video Memory\n","");break;
			case DDERR_NODIRECTDRAWHW:			debug.Output ("<WinDX::CreateSurface> No Direct Draw HW\n","");break;
			case DDERR_NOCOOPERATIVELEVELSET:	debug.Output ("<WinDX::CreateSurface> No Cooperate Level Set\n","");break;
			case DDERR_INVALIDCAPS:				debug.Output ("<WinDX::CreateSurface> Invalid Caps\n","");break;
			case DDERR_INVALIDPIXELFORMAT:		debug.Output ("<WinDX::CreateSurface> Invalid Pixel Format\n","");break;
			case DDERR_NOALPHAHW:				debug.Output ("<WinDX::CreateSurface> No Alpha HW\n","");break;
			case DDERR_NOFLIPHW:				debug.Output ("<WinDX::CreateSurface> No Flip HW\n","");break;
			case DDERR_NOZBUFFERHW:				debug.Output ("<WinDX::CreateSurface> No Z-Buffer HW\n","");break;
			case DDERR_NOEXCLUSIVEMODE:			debug.Output ("<WinDX::CreateSurface> No Exclusive Mode\n","");break;
			case DDERR_OUTOFMEMORY: 			debug.Output ("<WinDX::CreateSurface> Out of Memory\n","");break;
			case DDERR_PRIMARYSURFACEALREADYEXISTS:	debug.Output ("<WinDX::CreateSurface> Primary Surface Already Exists\n","");break;
			case DDERR_NOEMULATION:				debug.Output ("<WinDX::CreateSurface> No Emulation\n","");		   break;
			case DDERR_INCOMPATIBLEPRIMARY:		debug.Output ("<WinDX::CreateSurface> Incompatible Primary\n",""); break;
			}
			ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
			status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL);		
		}			
		// No good, return error
		if (status!=DD_OK) {
			win_status = GAMEX_FRONTSURFACE_FAILED;			
			return false;
		}
		win_status = GAMEX_FRONTSURFACE_OK;
		// Set back surface description
	   	DDSCAPS caps;		
		caps.dwCaps = DDSCAPS_BACKBUFFER;
		status = dd_front_buf->GetAttachedSurface (&caps, &dd_back_buf);
		if (status!=DD_OK) {
			win_status = GAMEX_BACKSURFACE_FAILED;
			return false;
		}
		win_status = GAMEX_BACKSURFACE_OK;
	} else {		
		// Set front surface description
		ddsd.dwSize = sizeof (ddsd);
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_VIDEOMEMORY;
		// Try to create primary surface 
		status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL);
		// No good, try using system memory
		if (status!=DD_OK) {			
			ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
			status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL);
		}
		// No good, return error
		if (status!=DD_OK) {
			win_status = GAMEX_FRONTSURFACE_FAILED;
			return false;
		}	
		win_status = GAMEX_FRONTSURFACE_OK;
		// Set back surface description
		ddsd.dwSize = sizeof (ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		ddsd.dwWidth = win_width;
		ddsd.dwHeight = win_height;
		status = dd_main->CreateSurface (&ddsd, &dd_back_buf, NULL);
		// No good, try using system memory
		if (status!=DD_OK) {
			ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
			status = dd_main->CreateSurface (&ddsd, &dd_back_buf, NULL);
		}
		// No good, return error
		if (status!=DD_OK) {
			win_status = GAMEX_BACKSURFACE_FAILED;
			return false;
		}
		win_status = GAMEX_BACKSURFACE_OK;
	} 

    // Get Surface Description (pixel format, pitch, etc.)	
	ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH | DDSD_PITCH | DDSD_PIXELFORMAT;
 
	dd_front_buf->GetSurfaceDesc(&ddsd);
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY) {
		debug.Output ("WinDX:: Front Buffer is in video memory.\n", "");
	} else {
		debug.Output ("WinDX:: Front Buffer NOT IN VIDEO MEMORY.\n", "");
	}
	
	dd_back_buf->GetSurfaceDesc(&ddsd);
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY) {		
		debug.Output ("WinDX:: Back Buffer is in video memory.\n", "");
	} else {
		debug.Output ("WinDX:: Back Buffer NOT IN VIDEO MEMORY.\n", "");
	}
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_BACKBUFFER) {
		debug.Output ("WinDX:: Back Buffer is a valid flip (back) buffer.\n", "");
	} else {
		debug.Output ("WinDX:: Back Buffer NOT A FLIP (BACK) BUFFER.\n", "");
	}	

	// Initialize Drawing Functions, Pixel Masks, Pitch and Width 
	// based on Surface Description	
	win_pitch = ddsd.lPitch;
	win_height = ddsd.dwHeight;	
	switch (win_screenbpp) {
	case 16: {		
		win_width = win_pitch / 2;
		FuncDrawPixel = &DrawPixel16;
		FuncDrawImage = &DrawImage16;
		FuncDrawMasked = &DrawMasked16;
		FuncDrawAlpha = &DrawAlpha16;
		FuncDrawAdded = &DrawAdded16;
		
		win_maskr = ddsd.ddpfPixelFormat.dwRBitMask;
		win_maskg = ddsd.ddpfPixelFormat.dwGBitMask; 		
		win_maskb = ddsd.ddpfPixelFormat.dwBBitMask;
		
		win_shiftr = GetShift (win_maskr);
		win_shiftg = GetShift (win_maskg);		
		win_shiftb = GetShift (win_maskb);

		win_maxr = GetMax (win_maskr);
		win_maxg = GetMax (win_maskg);
		win_maxb = GetMax (win_maskb);
	} break;
	case 24: {
		win_width = win_pitch / 3;
		FuncDrawPixel = &DrawPixel16;
		FuncDrawImage = &DrawImage24;
		FuncDrawMasked = &DrawMasked24;
		FuncDrawAlpha = &DrawAlpha24;
		FuncDrawAdded = &DrawAdded16;		
		win_maskr = ddsd.ddpfPixelFormat.dwRBitMask;
		win_maskg = ddsd.ddpfPixelFormat.dwGBitMask; 
		win_maskb = ddsd.ddpfPixelFormat.dwBBitMask;		
		win_shiftr = 16;
		win_shiftg = 8;
		win_shiftb = 0;
		win_maxr = 8;
		win_maxg = 8;
		win_maxb = 8;		
	} break;
	default: {
		win_width = ddsd.dwWidth;
		FuncDrawPixel = NULL;
		FuncDrawImage = NULL;
		FuncDrawMasked = NULL;
		FuncDrawAlpha = NULL;
		FuncDrawAdded = NULL;
	} break;
	}



	// Precompute: y-table look up 
	for (int y=0; y<=win_height; y++)
		win_ytable[y] = y*win_pitch;
		
	// Precompute: screen size (total number of bytes)
	win_size = win_width*win_height;
	if (win_screenbpp==16)	win_size <<= 1;	
	
	// Precompute: client rectangle
	
	GetClientRect(win_hwnd, &win_destrect);
    pt.x = pt.y = 0;
    ClientToScreen(win_hwnd, &pt);
	OffsetRect(&win_destrect, pt.x, pt.y);
	win_xpos = pt.x; win_ypos = pt.y;
	
	win_active = true;	
	win_status = GAMEX_READY;
    return true;
}

bool WindowsDX::DirectDrawCleanUp (void)
{
	// Direct Draw 8.1
	/*
	if (dd_device) dd_device->Release();
    if (dd_main) dd_main->Release();	
	if (dd_caps) delete dd_caps; */

	// Direct Draw 3.0
	if (dd_back_buf)	dd_back_buf->Release (),	dd_back_buf = NULL;
	if (dd_front_buf)	dd_front_buf->Release (),	dd_front_buf = NULL;	
	if (dd_main)		dd_main->Release (),		dd_main = NULL;
	return true;
}

bool WindowsDX::CleanUp (void)
{
	if (DirectDrawCleanUp()==FALSE)	{
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::CleanUp: Unable to close Direct Draw.\n", "");
		#endif			
		return false;
	}
	return true;
}

int WindowsDX::ViewUpdate (void)
{
    // Direct X 8.1
	//dd_device->Present (NULL, NULL, win_hwnd, NULL);

	// Direct Draw 3.0
	HRESULT             status;	
	DDBLTFX				bltfx;	
	
	bltfx.dwSize = sizeof (DDBLTFX);
	bltfx.dwDDFX = DDBLTFX_NOTEARING;
	
	
    if (win_active==TRUE) {
		if (GetForegroundWindow() == win_hwnd) {
			if (win_fullscreen==TRUE) {
				dd_front_buf->Flip(NULL, DDFLIP_WAIT);	// Flip back buffer to front
				if (dd_front_buf->IsLost() == DDERR_SURFACELOST) 
					dd_front_buf->Restore();
			} else {				
				status = dd_front_buf->BltFast (win_xpos, win_ypos, dd_back_buf, 0, DDBLTFAST_WAIT);
				if (status==DD_OK) return TRUE;
				if (status==DDERR_WASSTILLDRAWING) return FALSE;
				if (status==DDERR_SURFACELOST) {
					MessageBox (win_hwnd, "<WinDX::ViewUpdate> Error: Surface Lost", "",MB_OK);
					status = dd_front_buf->Restore ();
					if (status!=DD_OK) return FALSE;
					status = dd_back_buf->Restore ();
					if (status!=DD_OK) return FALSE;				
				}				
			}			
		}		
	}
	return TRUE;
}


inline void WindowsDX::SetActivation (BOOL act)
{
	win_active = act;
}

inline void WindowsDX::SetPosition (int x, int y)
{
	win_xpos = x;
	win_ypos = y;
}

void WindowsDX::SetView (int x1, int y1, int x2, int y2)
{
	win_viewx1 = x1;
	win_viewy1 = y1;
	win_viewx2 = x2;
	win_viewy2 = y2;
}

inline void WindowsDX::SetKey (char ch)
{
	win_keybuf[win_keys++] = ch;
	if (win_keys > WINDX_BUFSIZE) {
		win_keys = WINDX_BUFSIZE;
		debug.Output ("WinDX: Keyboard buffer overflow.\n", "");
	}
}

char WindowsDX::GetKey (void) 
{	
	if (win_keys>0) {
		win_keys--;		
		return win_keybuf[win_keys];
    } else {
		return 0;
	}	
}

bool WindowsDX::DrawOpen (void)
{
	/*HRESULT status;

	status = dd_device->BeginScene();
	if (status!=D3D_OK) exit (-21); */

	// Direct Draw 3.0
	HRESULT				status;
	DDSURFACEDESC		ddsd;

	ddsd.dwSize = sizeof(ddsd);
	status = dd_back_buf->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);	// lock drawing surface
	if (status!=DD_OK) {
		win_status = GAMEX_LOCK_FAILED;
		ReportProblem();
		return false;
	}

	win_screen = (char *) ddsd.lpSurface;
	return true;
}

int WindowsDX::DrawClose (void)
{
	// Direct X 8.1
	//dd_device->EndScene();

	// Direct Draw 3.0
	dd_back_buf->Unlock (NULL);							// unlock drawing surface
	return TRUE;
}

void WindowsDX::DirectDrawError (int status)
{
	switch (status) {
	case DDERR_ALREADYINITIALIZED: debug.Output ("DirectX: This object is already initialized.\n",""); break;
	case DDERR_BLTFASTCANTCLIP: debug.Output ("DirectX: Return if a clipper object is attached to the source surface passed into a BltFast call.\n",""); break;
	case DDERR_CANNOTATTACHSURFACE: debug.Output ("DirectX: This surface can not be attached to the requested surface.\n",""); break; 
	case DDERR_CANNOTDETACHSURFACE: debug.Output ("DirectX: This surface can not be detached from the requested surface.\n","");break;
	case DDERR_CANTCREATEDC: debug.Output ("DirectX: Windows can not create any more DCs\n",""); break;
	case DDERR_CANTDUPLICATE: debug.Output ("DirectX: Can't duplicate primary & 3D surfaces, or surfaces that are implicitly created.\n",""); break;
	case DDERR_CLIPPERISUSINGHWND: debug.Output ("DirectX: An attempt was made to set a cliplist for a clipper object that is already monitoring an hwnd.\n",""); break;
	case DDERR_COLORKEYNOTSET: debug.Output ("DirectX: No src color key specified for this operation.\n",""); break;
	case DDERR_CURRENTLYNOTAVAIL: debug.Output ("DirectX: Support is currently not available.\n",""); break;
	case DDERR_DIRECTDRAWALREADYCREATED: debug.Output ("DirectX: A DirectDraw object representing this driver has already been created for this process.\n",""); break;
	case DDERR_EXCEPTION: debug.Output ("DirectX: An exception was encountered while performing the requested operation.\n",""); break;
	case DDERR_EXCLUSIVEMODEALREADYSET: debug.Output ("DirectX: An attempt was made to set the cooperative level when it was already set to exclusive.\n",""); break;
	case DDERR_GENERIC: debug.Output ("DirectX: Generic failure.\n",""); break;
	case DDERR_HEIGHTALIGN: debug.Output ("DirectX: Height of rectangle provided is not a multiple of reqd alignment.\n",""); break;
	case DDERR_HWNDALREADYSET: debug.Output ("DirectX: The CooperativeLevel HWND has already been set. It can not be reset while the process has surfaces or palettes created.\n",""); break;
	case DDERR_HWNDSUBCLASSED: debug.Output ("DirectX: HWND used by DirectDraw CooperativeLevel has been subclassed, this prevents DirectDraw from restoring state.\n",""); break;
	case DDERR_IMPLICITLYCREATED: debug.Output ("DirectX: This surface can not be restored because it is an implicitly created surface.\n",""); break;
	case DDERR_INCOMPATIBLEPRIMARY: debug.Output ("DirectX: Unable to match primary surface creation request with existing primary surface.\n",""); break;
	case DDERR_INVALIDCAPS: debug.Output ("DirectX: One or more of the caps bits passed to the callback are incorrect.\n",""); break;
	case DDERR_INVALIDCLIPLIST: debug.Output ("DirectX: DirectDraw does not support the provided cliplist.\n",""); break;
	case DDERR_INVALIDDIRECTDRAWGUID: debug.Output ("DirectX: The GUID passed to DirectDrawCreate is not a valid DirectDraw driver identifier.\n",""); break;
	case DDERR_INVALIDMODE: debug.Output ("DirectX: DirectDraw does not support the requested mode.\n",""); break;
	case DDERR_INVALIDOBJECT: debug.Output ("DirectX: DirectDraw received a pointer that was an invalid DIRECTDRAW object.\n",""); break;
	case DDERR_INVALIDPARAMS: debug.Output ("DirectX: One or more of the parameters passed to the function are incorrect.\n",""); break;
	case DDERR_INVALIDPIXELFORMAT: debug.Output ("DirectX: The pixel format was invalid as specified.\n",""); break;
	case DDERR_INVALIDPOSITION: debug.Output ("DirectX: Returned when the position of the overlay on the destination is no longer legal for that destination.\n",""); break;
	case DDERR_INVALIDRECT: debug.Output ("DirectX: Rectangle provided was invalid.\n",""); break;
	case DDERR_LOCKEDSURFACES: debug.Output ("DirectX: Operation could not be carried out because one or more surfaces are locked.\n",""); break;
	case DDERR_NO3D: debug.Output ("DirectX: There is no 3D present.\n",""); break;
	case DDERR_NOALPHAHW: debug.Output ("DirectX: Operation could not be carried out because there is no alpha accleration hardware present or available.\n",""); break;
	//case DDERR_NOANTITEARHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support for synchronizing blits to avoid tearing.	\n",""); break;
	case DDERR_NOBLTHW: debug.Output ("DirectX: No blitter hardware present.\n",""); break;
	//case DDERR_NOBLTQUEUEHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support for asynchronous blitting.\n",""); break;
	case DDERR_NOCLIPLIST: debug.Output ("DirectX: No cliplist available.\n",""); break;
	case DDERR_NOCLIPPERATTACHED: debug.Output ("DirectX: No clipper object attached to surface object.\n",""); break;
	case DDERR_NOCOLORCONVHW: debug.Output ("DirectX: Operation could not be carried out because there is no color conversion hardware present or available.\n",""); break;
	case DDERR_NOCOLORKEY: debug.Output ("DirectX: Surface doesn't currently have a color key\n",""); break;
	case DDERR_NOCOLORKEYHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support of the destination color key.\n",""); break;
	case DDERR_NOCOOPERATIVELEVELSET: debug.Output ("DirectX: Create function called without DirectDraw object method SetCooperativeLevel being called.\n",""); break;
	case DDERR_NODC: debug.Output ("DirectX: No DC was ever created for this surface.\n",""); break;
	case DDERR_NODDROPSHW: debug.Output ("DirectX: No DirectDraw ROP hardware.\n",""); break;
	case DDERR_NODIRECTDRAWHW: debug.Output ("DirectX: A hardware-only DirectDraw object creation was attempted but the driver did not support any hardware.\n",""); break;
	case DDERR_NOEMULATION: debug.Output ("DirectX: Software emulation not available.\n",""); break;
	case DDERR_NOEXCLUSIVEMODE: debug.Output ("DirectX: Operation requires the application to have exclusive mode but the application does not have exclusive mode.\n",""); break;
	case DDERR_NOFLIPHW: debug.Output ("DirectX: Flipping visible surfaces is not supported.\n",""); break;
	case DDERR_NOGDI: debug.Output ("DirectX: There is no GDI present.\n",""); break;
	case DDERR_NOHWND: debug.Output ("DirectX: Clipper notification requires an HWND or no HWND has previously been set as the CooperativeLevel HWND.\n",""); break;
	case DDERR_NOMIRRORHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware present or available.\n",""); break;
	case DDERR_NOOVERLAYDEST: debug.Output ("DirectX: Returned when GetOverlayPosition is called on an overlay that UpdateOverlay has never been called on to establish a destination.\n",""); break;
	case DDERR_NOOVERLAYHW: debug.Output ("DirectX: Operation could not be carried out because there is no overlay hardware present or available.\n",""); break;
	case DDERR_NOPALETTEATTACHED: debug.Output ("DirectX: No palette object attached to this surface.\n",""); break;
	case DDERR_NOPALETTEHW: debug.Output ("DirectX: No hardware support for 16 or 256 color palettes.\n",""); break;
	case DDERR_NORASTEROPHW: debug.Output ("DirectX: Operation could not be carried out because there is no appropriate raster op hardware present or available.\n",""); break;
	case DDERR_NOROTATIONHW: debug.Output ("DirectX: Operation could not be carried out because there is no rotation hardware present or available.\n",""); break;
	case DDERR_NOSTRETCHHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support for stretching.\n",""); break;
	case DDERR_NOT4BITCOLOR: debug.Output ("DirectX: DirectDrawSurface is not in 4 bit color palette and the requested operation requires 4 bit color palette.\n",""); break;
	case DDERR_NOT4BITCOLORINDEX: debug.Output ("DirectX: DirectDrawSurface is not in 4 bit color index palette and the requested operation requires 4 bit color index palette.\n",""); break;
	case DDERR_NOT8BITCOLOR: debug.Output ("DirectX: DirectDrawSurface is not in 8 bit color mode and the requested operation requires 8 bit color.\n",""); break;
	case DDERR_NOTAOVERLAYSURFACE: debug.Output ("DirctX: Returned when an overlay member is called for a non-overlay surface.\n",""); break;
	case DDERR_NOTEXTUREHW: debug.Output ("DirectX: Operation could not be carried out because there is no texture mapping hardware present or available.\n",""); break;
	case DDERR_NOTFLIPPABLE: debug.Output ("DirectX: An attempt has been made to flip a surface that is not flippable.\n",""); break;
	case DDERR_NOTFOUND: debug.Output ("DirectX: Requested item was not found.\n",""); break;
	case DDERR_NOTLOCKED: debug.Output ("DirectX: Surface was not locked.  An attempt to unlock a surface that was not locked at all, or by this process, has been attempted.\n",""); break;
	case DDERR_NOTPALETTIZED: debug.Output ("DirectX: The surface being used is not a palette-based surface.\n",""); break;
	case DDERR_NOVSYNCHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support for vertical blank synchronized operations.\n",""); break;
	case DDERR_NOZBUFFERHW: debug.Output ("DirectX: Operation could not be carried out because there is no hardware support for zbuffer blitting.\n",""); break;
	case DDERR_NOZOVERLAYHW: debug.Output ("DirectX: Overlay surfaces could not be z layered based on their BltOrder because the hardware does not support z layering of overlays.\n",""); break;
	case DDERR_OUTOFCAPS: debug.Output ("DirectX: The hardware needed for the requested operation has already been allocated.\n",""); break;
	case DDERR_OUTOFMEMORY: debug.Output ("DirectX: DirectDraw does not have enough memory to perform the operation.\n",""); break;
	case DDERR_OUTOFVIDEOMEMORY: debug.Output ("DirectX: DirectDraw does not have enough memory to perform the operation.\n",""); break;
	case DDERR_OVERLAYCANTCLIP: debug.Output ("DirectX: The hardware does not support clipped overlays.\n",""); break;
	case DDERR_OVERLAYCOLORKEYONLYONEACTIVE: debug.Output ("DirectX: Can only have ony color key active at one time for overlays.\n",""); break;
	case DDERR_OVERLAYNOTVISIBLE: debug.Output ("DirectX: Returned when GetOverlayPosition is called on a hidden overlay.\n",""); break;
	case DDERR_PALETTEBUSY: debug.Output ("DirectX: Access to this palette is being refused because the palette is already locked by another thread.\n",""); break;
	case DDERR_PRIMARYSURFACEALREADYEXISTS: debug.Output ("DirectX: This process already has created a primary surface.\n",""); break;
	case DDERR_REGIONTOOSMALL: debug.Output ("DirectX: Region passed to Clipper::GetClipList is too small.\n",""); break;
	case DDERR_SURFACEALREADYATTACHED: debug.Output ("DirectX: This surface is already attached to the surface it is being attached to.\n",""); break;
	case DDERR_SURFACEALREADYDEPENDENT: debug.Output ("DirectX: This surface is already a dependency of the surface it is being made a dependency of.\n",""); break;
	case DDERR_SURFACEBUSY: debug.Output ("DirectX: Access to this surface is being refused because the surface is already locked by another thread.\n",""); break;
	case DDERR_SURFACEISOBSCURED: debug.Output ("DirectX: Access to surface refused because the surface is obscured.\n",""); break;
	case DDERR_SURFACELOST: debug.Output ("DirectX: Access to this surface is being refused because the surface memory is gone. The DirectDrawSurface object representing this surface should have Restore called on it.\n",""); break;
	case DDERR_SURFACENOTATTACHED: debug.Output ("DirectX: The requested surface is not attached.\n",""); break;
	case DDERR_TOOBIGHEIGHT: debug.Output ("DirectX: Height requested by DirectDraw is too large.\n",""); break;
	case DDERR_TOOBIGSIZE: debug.Output ("DirectX: Size requested by DirectDraw is too large --	the individual height and width are OK.\n",""); break;
	case DDERR_TOOBIGWIDTH: debug.Output ("DirectX: Width requested by DirectDraw is too large.\n",""); break;
	case DDERR_UNSUPPORTED: debug.Output ("DirectX: Action not supported.\n",""); break;
	case DDERR_UNSUPPORTEDFORMAT: debug.Output ("DirectX: FOURCC format requested is unsupported by DirectDraw.\n",""); break;
	case DDERR_UNSUPPORTEDMASK: debug.Output ("DirectX: Bitmask in the pixel format requested is unsupported by DirectDraw.\n",""); break;
	case DDERR_VERTICALBLANKINPROGRESS: debug.Output ("DirectX: Vertical blank is in progress.\n",""); break;
	case DDERR_WASSTILLDRAWING: debug.Output ("DirectX: Informs DirectDraw that the previous Blt which is transfering information to or from this Surface is incomplete.\n",""); break;
	case DDERR_WRONGMODE: debug.Output ("DirectX: This surface can not be restored because it was created in a different mode.\n",""); break;
	}
}
void WindowsDX::DrawClear (void)
{	
	unsigned short *start;
	int xres;
	start = (unsigned short *) win_screen;
	xres = win_pitch;
	for (int y=0; y<win_height; y++) {			
		memset (start, 0, xres);
		start += win_width;		
	}
}



void WindowsDX::DrawColor (int r, int g, int b)
{
	win_r = r;	win_g = g;	win_b = b;
	win_rgb = RGB(win_r, win_g, win_b);
}

bool WindowsDX::ClipImage (int x, int y, ImageX *img, int &view_offset, int &img_offset, int &img_xs, int &img_ys)
{
	int x2, y2;

	img_xs = img->m_xres;
	img_ys = img->m_yres;
	x2 = x+(img_xs-1);
	y2 = y+(img_ys-1);
	
	if (x2 >= win_viewx1 && x <= win_viewx2) {			// In-View Horizontally
		if (y2 >= win_viewy1 && y <= win_viewy2) {		// In-View Vertically 
			view_offset = win_viewy1*win_width + win_viewx1;
			img_offset = 0;			
			if (x < win_viewx1) {						// Clip Left
				img_xs -= (win_viewx1 - x);
				img_offset += (win_viewx1 - x);
			} else {
				view_offset += (x - win_viewx1);
			}
			if (x2 > win_viewx2)						// Clip Right
				img_xs -= (x2 - win_viewx2);
			if (y < win_viewy1) {						// Clip Top
				img_ys -= (win_viewy1 - y);
				img_offset += (win_viewy1 - y) * img->m_xres;	
			} else {
				view_offset += (y - win_viewy1) * win_width;
			}
			if (y2 > win_viewy2)						// Clip Bottom
				img_ys -= (y2 - win_viewy2);
			return true;
		}
	}
	return false;			
}

void WindowsDX::UseAlpha (void)
{
	int a, b, c;	
	table_alpha = new unsigned short[32*32];
	
	for (a=0; a<32; a++) {
		for (b=0; b<32; b++) {
			c = a+b; if (c>31) c = 31;
			table_alpha[(a<<5) + b] = c;			
			//table_alpha[(a<<5 + b)<<5] = c<<5;
			//table_alpha[a<<5 + b] = c;
		}
	}
}

void WindowsDX::DrawBox (int x1, int y1, int x2, int y2, int r, int g, int b)
{ 
	unsigned short *start, *pixel;
	int i, j, clr;

	clr = ((r>>3) << 10) + ((g>>3) << 5) + (b>>3);
	start = (unsigned short *) win_screen;
	start += y1*win_pitch + x1;
	for (j=y1; j<=y2; j++) {
		pixel = start;
		for (i=x1; i<=x2; i++)
			*pixel++ = clr;
		start += win_pitch;
	}
}

int WindowsDX::GetWidth ()
{
	return win_width;	
}

int WindowsDX::GetHeight ()
{
	return win_height;
}

int WindowsDX::GetPitch ()
{
	return win_pitch;
}

void *WindowsDX::GetScreen (void)
{
	return win_screen;
}

HWND WindowsDX::GetWindow (void)
{
	return win_hwnd;
}

int WindowsDX::DrawText (int x, int y, char *msg)
{
	HRESULT hr;
    HDC hdc = 0;
    
	hr = dd_back_buf->GetDC(&hdc);		// Get the device context handle.
    if (FAILED(hr)) { 
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::DrawText: Unable to get device context handle.\n", "");
		#endif	
		return FALSE; 
	}
										// Write the message.
	SetBkMode(hdc, TRANSPARENT);
	SetTextColor(hdc, win_rgb);
    TextOut(hdc, x, y, msg, strlen(msg));
										// Release the device context.
	hr = dd_back_buf->ReleaseDC(hdc);
	return TRUE;
}

void WindowsDX::SetMouseMove (int x, int y)
{
	mouse_xl = mouse_x;
	mouse_yl = mouse_y;
	mouse_x = x;
	mouse_y = y;
	mouse_motion = TRUE;
}

void WindowsDX::SetMousePress (int button)
{
	mouse_down = TRUE;
	mouse_button = button;
}

void WindowsDX::SetMouseRelease ()
{
	mouse_down = FALSE;
	mouse_button = FALSE;
}

int WindowsDX::GetMousePress ()
{
	return mouse_button;
}

int WindowsDX::GetMouseX ()
{
	return mouse_x;
}

int WindowsDX::GetMouseY ()
{
	return mouse_y;
}

float WindowsDX::GetMouseDX ()
{
	return (float) ((mouse_x - mouse_xl)/MOUSE_POS_RATE);
}

float WindowsDX::GetMouseDY ()
{
	return (float) ((mouse_y - mouse_yl)/MOUSE_POS_RATE);
}

void WindowsDX::ClipTile (int x, int y, int xres, int yres, int &mx, int &my, int &offx, int &offy)
{
	if (x<0) {
		mx = ((int) (x / xres)) - 1;
		offx = (x % xres) + xres;
	} else {
		mx = ((int) (x / xres));
		offx = x % xres;
	}
	if (y<0) {
		my = ((int) (y / yres)) - 1;		
		offy = (y % yres) + yres;
	} else {
		my = ((int) (y / yres));		
		offy = y % yres;
	}	
}

int WindowsDX::DrawBackground (int x, int y, ImageX *img)
{
	int xres, yres;
	int map_x1, map_y1;
	int map_x2, map_y2;	
	int map_numx, map_numy;	
	int off_x, off_y;
	int img_x, img_y;	
	int mx, my;

	xres = img->m_xres; yres = img->m_yres;
	ClipTile (x+win_width, y+win_height, xres, yres, map_x2, map_y2, off_x, off_y);
	ClipTile (x, y, xres, yres, map_x1, map_y1, off_x, off_y);

	map_numx = map_x2 - map_x1 + 1;
	map_numy = map_y2 - map_y1 + 1; 	
	for (my = 0; my < map_numy; my++) {
		img_y = my * yres - off_y;
		img_x = -off_x;
		for (mx = 0; mx < map_numx; mx++) {
			DrawImage (img_x, img_y, img);
			img_x += xres;
		}
	}
	return 1;
}



static int DrawPixel16 (int x, int y, int r, int g, int b)
{	
	*((unsigned short *) GameX.win_screen + GameX.win_ytable[y] + x) = (r<<10) + (g<<5) + b;
	return 1;
}

static int DrawImage16 (int x, int y, ImageX *img)
{
	XBYTE2 *img_start;
	unsigned short *view_start;
	int img_offset, view_offset;
	int img_xs, img_ys;
	int img_wid;
		
	if (img!=NULL && img->GetUsage()==DRAW_IMAGE) {
		img_start = (XBYTE2 *) img->m_data;
		if (img_start != NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
			view_start = (unsigned short *) GameX.win_screen + view_offset;
			img_start += img_offset;
			img_wid = img_xs << 1;
			for (y=0; y<img_ys; y++) {			
				memcpy (view_start, img_start, img_wid);
				view_start += GameX.win_width;
				img_start += img->m_xres;
			} 		
		}
		return 1;
	} else {		
		GameX.ReportProblem(GAMEX_IMAGEUSE_FAILED);
		return 0;
	}	
}

static int DrawMasked16 (int x, int y, ImageX *img)
{
	XBYTE2 *img_start;
	unsigned short *view_start;	
	int img_offset, view_offset;
	int img_xs, img_ys;
	XBYTE2 *img_pix;
	unsigned short *view_pix;
	
	if (img!=NULL && img->GetUsage()==DRAW_MASKED) {
		img_start = (XBYTE2 *) img->m_data;
		if (img_start != NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
			view_start = (unsigned short *) GameX.win_screen + view_offset;
			img_start += img_offset;
			for (y=0; y<img_ys; y++) {			
				img_pix = img_start;
				for (view_pix=view_start; view_pix<view_start+img_xs;) {
					if (*img_pix>0) *view_pix = *img_pix;
					view_pix++; img_pix++;
				}
				view_start += GameX.win_width;
				img_start += img->m_xres;
			} 		
		}
		return 1;
	} else {
		GameX.ReportProblem(GAMEX_IMAGEUSE_FAILED);
		return 0;
	}
}

static int DrawAlpha16 (int x, int y, ImageX *img)
{
	XBYTE2 *img_start;
	XBYTE *alpha_start;
	unsigned short *view_start;
	int img_offset, view_offset;
	int img_xs, img_ys;
	XBYTE2 *img_pix;
	XBYTE *alpha_pix;
	unsigned short *view_pix;
	register int i, v;
	
	if (img!=NULL && img->GetUsage()==DRAW_ALPHA) {
		img_start = (XBYTE2 *) img->m_data;
		alpha_start = (XBYTE *) img->m_alpha;
		if (img_start!=NULL && alpha_start!=NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
			view_start = (unsigned short *) GameX.win_screen + view_offset;
			img_start += img_offset;
			alpha_start += img_offset;
			for (y=0; y<img_ys; y++) {			
				img_pix = img_start;
				alpha_pix = alpha_start;
				for (view_pix=view_start; view_pix<view_start+img_xs;) {
					i = *alpha_pix;
					if (i!=255) {
						v = *view_pix;					
						*view_pix = ((v & win_maskb) * i) >> 8;
						*view_pix |= (((v & win_maskg) * i) >> 8) & win_maskg;
						*view_pix |= (((v & win_maskr) * i) >> 8) & win_maskr;
						*view_pix += *img_pix;
					}
					view_pix++; img_pix++; alpha_pix++;
				}
				view_start += GameX.win_width;
				img_start += img->m_xres;
				alpha_start += img->m_xres;
			} 		
		}
		return 1;
	} else {		
		GameX.ReportProblem(GAMEX_IMAGEUSE_FAILED);
		return 0;
	}	
}

static int DrawAdded16 (int x, int y, ImageX *img)
{
	XBYTE2 *img_start;
	XBYTE *alpha_start;
	unsigned short *view_start;
	int img_offset, view_offset;
	int img_xs, img_ys;
	XBYTE2 *img_pix;
	XBYTE *alpha_pix;
	unsigned short *view_pix;
	register int i, v, c;

	if (img!=NULL && img->GetUsage()==DRAW_ADDED) {
		img_start = (XBYTE2 *) img->m_data;
		alpha_start = (XBYTE *) img->m_alpha;
		if (img_start!=NULL && alpha_start!=NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
			view_start = (unsigned short *) GameX.win_screen + view_offset;
			img_start += img_offset;
			alpha_start += img_offset;
			for (y=0; y<img_ys; y++) {			
				img_pix = img_start;
				alpha_pix = alpha_start;
				for (view_pix=view_start; view_pix<view_start+img_xs;) {								
					v = *view_pix; i = *img_pix;
					if (v!=255 && i!=0) {
						c = (v & win_maskb) + (i & win_maskb); if (c>win_maskb) c = win_maskb;
						*view_pix = c;
						c = (v & win_maskg) + (i & win_maskg); if (c>win_maskg) c = win_maskg;
						*view_pix |= c;
						c = (v & win_maskr) + (i & win_maskr); if (c>win_maskr) c = win_maskr;
						*view_pix |= c;
					}
					view_pix++; img_pix++;
				}
				view_start += GameX.win_width;
				img_start += img->m_xres;
				alpha_start += img->m_xres;
			} 
		}
		return 1;
	} else {		
		GameX.ReportProblem(GAMEX_IMAGEUSE_FAILED);
		return 0;
	}

}

static int DrawImage24 (int x, int y, ImageX *img)
{
	XBYTE *img_start;
	XBYTE *view_start;
	int img_offset, view_offset;
	int img_xs, img_ys;
	int img_clipwid, view_wid, img_xwid;

	img_start = (XBYTE*) img->m_data;
	if (img_start != NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
		view_start = (XBYTE *) GameX.win_screen + view_offset*3;
		img_start += img_offset*3;
		img_clipwid = img_xs * 3;
		img_xwid = img->m_xres * 3;
		view_wid = GameX.win_width * 3;
		for (y=0; y<img_ys; y++) {			
			memcpy (view_start, img_start, img_clipwid);
			view_start += view_wid;
			img_start += img_xwid;
		} 		
	}
	return 1;
}

static int DrawMasked24 (int x, int y, ImageX *img)
{
	XBYTE *img_start;
	XBYTE *view_start;	
	int img_offset, view_offset;
	int img_xs, img_ys;
	XBYTE *img_pix;
	XBYTE *view_pix;
	int img_clipwid;
	int img_xwid, view_wid;
	
	img_start = (XBYTE *) img->m_data;
	if (img_start != NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
		view_start = (XBYTE *) GameX.win_screen + view_offset*3;
		img_start += img_offset;
		img_clipwid = img_xs * 3;
		img_xwid = img->m_xres * 3;
		view_wid = GameX.win_width * 3;
		for (y=0; y<img_ys; y++) {			
			img_pix = img_start;
			for (view_pix=view_start; view_pix<view_start+img_clipwid;) {
				if (*img_pix>0) {
					*view_pix++ = *img_pix++;
					*view_pix++ = *img_pix++;
					*view_pix++ = *img_pix++;
				} else {
					view_pix+=3; img_pix+=3;
				}				
			}
			view_start += view_wid;
			img_start += img_xwid;
		} 		
	}	
	return 1;
}

static int DrawAlpha24 (int x, int y, ImageX *img)
{
	XBYTE *img_start;
	XBYTE *alpha_start;
	XBYTE *view_start;
	int img_offset, view_offset;
	int img_xs, img_ys;
	XBYTE *img_pix;
	XBYTE *alpha_pix;
	XBYTE *view_pix;
	register int i;
	int img_clipwid, img_xwid, view_wid;
	
	img_start = (XBYTE*) img->m_data;
	alpha_start = (XBYTE*) img->m_alpha;
	if (img_start!=NULL && alpha_start!=NULL && GameX.ClipImage(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
		view_start = (XBYTE*) GameX.win_screen + view_offset*3;
		img_start += img_offset*3;
		alpha_start += img_offset*3;
		img_clipwid = img_xs * 3;
		img_xwid = img->m_xres * 3;
		view_wid = GameX.win_width * 3;
		for (y=0; y<img_ys; y++) {			
			img_pix = img_start;
			alpha_pix = alpha_start;
			for (view_pix=view_start; view_pix<view_start+img_clipwid;) {
				i = *alpha_pix;
				if (i!=255) {
					*view_pix++ = ((*view_pix * i) >> 8) + *img_pix;
					*view_pix++ = ((*view_pix * i) >> 8) + *img_pix;
					*view_pix++ = ((*view_pix * i) >> 8) + *img_pix;					
				} else {
					view_pix+=3; img_pix+=3;
				}
				alpha_pix++;
			}
			view_start += view_wid;
			img_start += img_xwid;
			alpha_start += img->m_xres;
		} 		
	}
	return 1;
}

void WindowsDX::ReportProblem (int problem)
{
	win_status = problem;
	ReportProblem ();
}

void WindowsDX::ReportProblem (void)
{	
	char error[100];
	char recommend[300];	
	char disp[1000];
	
	switch (win_status) {
	case GAMEX_NO_INIT: {
		sprintf (error, "Error: NO_INIT (User did not initialize GameX)");
		sprintf (recommend, "You must start GameX by calling:\n  GameX.Initialize (fullscreen, xres, yres)\n\nfrom the initialize() function. See documentation for starting GameX.");
	} break;
	case GAMEX_INITCLASS_FAILED: {
		sprintf (error, "Error: INITCLASS_FAILED (Cannot Initialize Window Class)");
		sprintf (recommend, "Are you using Windows 98/2000/NT? GameX requires one of these. Still having trouble? Please report error and system configuration.");
	} break;
	case GAMEX_INITWINTEMP_FAILED: {
		sprintf (error, "Error: INITWINTEMP_FAILED (Cannot Initialize Temp Window)");
		sprintf (recommend, "Are you using Windows 98/2000/NT? GameX requires one of these. Still having trouble? Please report error and system configuration.");
	} break;
	case GAMEX_INITWINREAL_FAILED: {
		sprintf (error, "Error: INITWINREAL_FAILED (Cannot Initialize Main Window)");
		sprintf (recommend, "Are you using Windows 98/2000/NT? GameX requires one of these. Still having trouble? Please report error and system configuration.");
	} break;
	case GAMEX_SHOWWIN_FAILED: {
		sprintf (error, "Error: SHOWWIN_FAILED (Unable to show Main Window)");
		sprintf (recommend, "Are you using Windows 98/2000/NT? GameX requires one of these. Still having trouble? Please report error and system configuration.");
	} break;
	case GAMEX_DDSTART_FAILED: {
		sprintf (error, "Error: DDSTART_FAILED (Unable to start DirectX)");
		sprintf (recommend, "Are you using DirectX SDK 3.1 or 8.1? GameX requires one of these. Still having trouble? Please report error and system configuration.");
	} break;
	case GAMEX_DDSETLEVEL_FAILED: {
		sprintf (error, "Error: DDSETLEVEL_FAILED (Unable to set DirectX Cooperative Level)");
		sprintf (recommend, "Are you using DirectX SDK 3.1 or 8.1? GameX requires one of these. Still having trouble? Please report error and system configuration.");
	} break;
	case GAMEX_SETMODE_FAILED: {
		sprintf (error, "Error: SETMODE_FAILED (Unable to set screen mode)");
		sprintf (recommend, "Are you using DirectX SDK 3.1 or 8.1? GameX requires one of these. Still having trouble? Please report error and system configuration.");
	} break;
	case GAMEX_FRONTSURFACE_FAILED: {
		sprintf (error, "Error: FRONTSURFACE_FAILED (Unable to create primary surface)");
		sprintf (recommend, "Are you using DirectX SDK 3.1 or 8.1? GameX requires one of these. Still having trouble? Please report error and system configuration.");		
	} break;
	case GAMEX_BACKSURFACE_FAILED: {
		sprintf (error, "Error: BACKSURFACE_FAILED (Unable to create secondary surface)");
		sprintf (recommend, "Are you using DirectX SDK 3.1 or 8.1? GameX requires one of these. Still having trouble? Please report error and system configuration.");
	} break;
	case GAMEX_REQSIZE_FAILED: {
		sprintf (error, "Error: REQSIZE_FAILED (Requested size too large)");
		sprintf (recommend, "The requested game size must be smaller than the current screen settings. Reduce requested size to less than the screen size or use fullscreen mode.\n\nIf you are unsure of screen size initialize GameX as follows:\n\nGameX.Initialize (false, 700, 500);");
	} break;	
	case GAMEX_BPP_FAILED: {
		sprintf (error, "Error: BPP_FAILED (Screen depth not supported)");
		sprintf (recommend, "GameX supports 16-bit screen depths ONLY!\n\nPlease adjust your screen depth to 16-bit as follows:\n- Right click on desktop\n- Select Properties\n- Goto Settings Tab\n- Select High Color (16-bit)");
	} break;
	case GAMEX_SOUNDINIT_FAILED: {
		sprintf (error, "Error: SOUNDINIT_FAILED (Unable to initialize Direct Sound)");
		sprintf (recommend, "GameX unable to initialize sound. Is sound card installed and working?");
	} break;
	case GAMEX_SOUNDCOOP_FAILED: {
		sprintf (error, "Error: SOUNDCOOP_FAILED (Unable to set Direct Sound Cooperative Level");
		sprintf (recommend, "GameX unable to set Direct Sound cooperative level. Is sound card installed and working?");
	} break;	
	case GAMEX_SOUNDCREATE_FAILED: {
		sprintf (error, "Error: SOUNDCREATE_FAILED (Unable to create Direct Sound buffer.");
		sprintf (recommend, "GameX unable to create a sound buffer for your sound. Try shortening the sound file (total length), reducing the sound resolution (bits per sample), reducing the sound quality (samples per second), or eliminating a channel (mono vs. stereo).");
	} break;
	case GAMEX_SOUNDMATCH_FAILED: {
		sprintf (error, "Error: SOUNDMATCH_FAILED (Loaded sounds do not match.");
		sprintf (recommend, "You are attempting to load several sounds with different quality, resolution, and channel info. All sounds to be played simultaneously must have the same resolution (bits per sample), same quality (samples per second), and same number of channels (mono vs. stereo). Edit WAV files accordinging.");
	} break;
	case GAMEX_SOUNDCAPS_FAILED: {
		sprintf (error, "Error: SOUNDCAPS_FAILED (Unable to get Sound Buffer Capabilities.");
		sprintf (recommend, "GameX was unable to retrieve sound buffer capabilities. You may be running out of memory. Close any open applications. Make sure you have DirectX 3.1, 7 or 8.1.");
	} break;
	case GAMEX_SOUNDBUFLEN_FAILED: {
		sprintf (error, "Error: SOUNDBUFLEN_FAILED (Direct Sound buffer and SoundX buffers do not match.");
		sprintf (recommend, "Direct Sound and SoundX buffer lengths do not match. Internal error. Please report this problem and the code that generated it.");	
	} break;
	case GAMEX_SOUNDLOST_FAIL: {
		sprintf (error, "Error: SOUNDLOST_FAIL (Direct Sound buffer was lost.");
		sprintf (recommend, "Direct Sound buffer was lost. Make sure you are using DirectX 3.1, 7 or 8.1. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.");
	} break;
	case GAMEX_SOUNDINVALID_FAIL: {
		sprintf (error, "Error: SOUNDINVALID_FAIL (Direct Sound Invalid Call.");
		sprintf (recommend, "DirectX, via GameX, made an invalid sound call. Make sure you are using DirectX 3.1, 7 or 8.1. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.");
	} break;
	case GAMEX_SOUNDPARAM_FAIL: {
		sprintf (error, "Error: SOUNDPARAM_FAIL (Direct Sound Invalid Parameter.");
		sprintf (recommend, "DirectX, via GameX, used an invalid parameter. Make sure you are using DirectX 3.1, 7 or 8.1. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.");
	} break;
	case GAMEX_SOUNDPRIOLEVEL_FAIL: {
		sprintf (error, "Error: SOUNDPARAM_FAIL (Direct Sound needs Priority Level");
		sprintf (recommend, "DirectX, via GameX, is using an invalid priority level. Make sure you are using DirectX 3.1, 7 or 8.1. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.");		
	} break;	
	case GAMEX_IMAGEUSE_FAILED: {
		sprintf (error, "Error: IMAGEUSE_FAILED (Image used incorrectly)");
		sprintf (recommend, "Purpose of image on load does not match its use. Images loaded with DRAW_ALPHA must be drawn with DrawAlpha. Images loaded with DRAW_MASKED must be drawn with DrawMasked. Etc.");		
	} break;
	default: {
		sprintf (error, "Error: UNREGISTERED");
		sprintf (recommend, "Unregistered error with GameX. Please report.");
	} break;
	}
	sprintf (disp, "Unable to start GameX.\n\n%s\n\n%s", error, recommend);	
	MessageBox (NULL, disp, "GameX", MB_OK);	
	
	sprintf (disp, "GameX %s\n", error);
	debug.Output (disp, "");
}

void WindowsDX::PlaySound (SoundX* snd, int how)
{		
	if (sound_support) {
		if (snd->NeedUpdate()==true) UpdateSound (snd);	
		int num = snd->GetIndex();
	
		if (how==PLAY_REWIND) ds_sound_list[num]->SetCurrentPosition (0);	
		ds_sound_list[num]->Play (0,0,0);
	}
}

void WindowsDX::PlaySound (SoundX* snd, int how, float vol, float pan)
{		
	if (sound_support) {
		if (snd->NeedUpdate()==true) UpdateSound (snd);	
		int num = snd->GetIndex();
	
		if (how==PLAY_REWIND) ds_sound_list[num]->SetCurrentPosition (0);
		ds_sound_list[num]->SetVolume ((long) (-5000.0 + vol*5000.0));
		ds_sound_list[num]->SetPan ((long) (pan*10000.0));
		ds_sound_list[num]->Play (0,0,0);
	}
}

void WindowsDX::RewindSound (SoundX* snd)
{
	if (sound_support) {
		if (snd->NeedUpdate()==true) UpdateSound (snd);	
		int num = snd->GetIndex();	
		ds_sound_list[num]->SetCurrentPosition(0);
	}
}


void WindowsDX::StopSound (SoundX* snd)
{
	if (sound_support) {
		int num = snd->GetIndex();
		if (num>=0) ds_sound_list[num]->Stop ();
	}
}

void WindowsDX::UpdateSound (SoundX* snd)
{
	signed short *buf_a, *buf_b;	
	unsigned long bytes_a, bytes_b;	

	int num_samples, num_channels;
	int result, num;
	int n, c;
	
	// Determine Direct Sound buffer to use based on SoundX object.
	// (If Direct Sound buffer does not exist for SoundX object, create one)
	num = snd->GetIndex();
	if (num==-1) {
		num = AddSound (snd);
		snd->SetIndex (num);
	}
	
	// Get requested Direct Sound buffer capabilities
	DSBCAPS caps;
	ZeroMemory(&caps, sizeof(caps));
	caps.dwSize = sizeof(caps);
	if (DS_OK != ds_sound_list[num]->GetCaps (&caps)) ReportProblem(GAMEX_SOUNDCAPS_FAILED);
			
	// Get and check number of samples (SoundX sample count should match DS buffer count)
	num_channels = snd->GetNumChannels();
	num_samples = caps.dwBufferBytes / num_channels;
	if (num_samples!=snd->GetNumSamples()) ReportProblem(GAMEX_SOUNDBUFLEN_FAILED);
	
	// Lock Direct Sound buffer
	result = ds_sound_list[num]->Lock (0, caps.dwBufferBytes, (void**) &buf_a, &bytes_a, (void**) &buf_b, &bytes_b, 0);
	n = 0;
	while (result==DSERR_BUFFERLOST && n<10) {
        ds_sound_list[num]->Restore();
        result = ds_sound_list[num]->Lock (0, caps.dwBufferBytes, (void**) &buf_a, &bytes_a, (void**) &buf_b, &bytes_b, 0);
		n++;
    }
	// Check for possible buffer errors
	if (DS_OK!=result) {		
		if (result==DSERR_BUFFERLOST)			win_status = GAMEX_SOUNDLOST_FAIL;	
		else if (result==DSERR_INVALIDCALL)		win_status = GAMEX_SOUNDINVALID_FAIL;
		else if (result==DSERR_INVALIDPARAM)	win_status = GAMEX_SOUNDPARAM_FAIL;
		else if (result==DSERR_PRIOLEVELNEEDED)	win_status = GAMEX_SOUNDPRIOLEVEL_FAIL;
		ReportProblem();
	}
	// Copy SoundX data to Direct Sound buffer
	if (snd->GetBPS()<=8) {
		unsigned char *data = (unsigned char*) snd->GetData();
		unsigned char *curr_a = (unsigned char*) buf_a;
		for (n = 0; n < num_samples; n++)
			for (c=0; c < num_channels; c++)
				*curr_a++ = *data++;
	} else {
		signed short *data = (signed short*) snd->GetData();
		signed short *curr_a = buf_a;
		curr_a = buf_a;
		for (n = 0; n < num_samples; n++)
			for (c=0; c < num_channels; c++) 
				*curr_a++ = *data++;			
	}
	
	// Unlock Direct Sound buffer
	ds_sound_list[num]->Unlock (&buf_a, bytes_a, &buf_b, bytes_b);

	snd->SetRefresh (false);
}

int WindowsDX::AddSound (SoundX* snd)
{
	SoundXPtr*				new_list;
	LPDIRECTSOUNDBUFFER*	new_ds_list;
	
	if (num_sounds==max_sounds) {
		max_sounds += 10;
		new_list = new SoundXPtr[max_sounds];
		new_ds_list = new LPDIRECTSOUNDBUFFER[max_sounds];
		memcpy (new_list, sound_list, sizeof(SoundXPtr)*num_sounds);
		memcpy (new_ds_list, ds_sound_list, sizeof(LPDIRECTSOUNDBUFFER)*num_sounds);
		if (sound_list!=NULL) {delete [] sound_list;}
		if (ds_sound_list!=NULL) {delete [] ds_sound_list;}
		sound_list = new_list;
		ds_sound_list = new_ds_list;
	}	
	sound_list[num_sounds] = snd;
	
	if (num_sounds==0) {
		ds_num_channels = snd->GetNumChannels();
		ds_bits_per_sample = snd->GetBPS();
		ds_samples_per_sec = snd->GetSamplesPerSec();
	} else {
		if (snd->GetNumChannels()!=ds_num_channels) ReportProblem (GAMEX_SOUNDMATCH_FAILED);
		if (snd->GetBPS()!=ds_bits_per_sample) ReportProblem (GAMEX_SOUNDMATCH_FAILED);
		if (snd->GetSamplesPerSec()!=ds_samples_per_sec) ReportProblem (GAMEX_SOUNDMATCH_FAILED);
	}

	ds_sound_list[num_sounds] = NULL;
	
	WAVEFORMATEX format; 
	ZeroMemory(&format,sizeof(format)); 		
	format.wFormatTag = WAVE_FORMAT_PCM;
	format.nChannels = ds_num_channels;
	format.nSamplesPerSec = ds_samples_per_sec;
	format.wBitsPerSample = ds_bits_per_sample;
	format.nBlockAlign = (format.wBitsPerSample / 8) * format.nChannels; 
	format.nAvgBytesPerSec = format.nSamplesPerSec * format.nBlockAlign; 	
	
	DSBUFFERDESC bufferDesc; 
	ZeroMemory(&bufferDesc,sizeof(bufferDesc)); 
	bufferDesc.dwSize = sizeof(bufferDesc); 
	bufferDesc.dwFlags = DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME; 
	bufferDesc.dwBufferBytes = snd->GetNumSamples() * format.nBlockAlign;
	bufferDesc.lpwfxFormat = (LPWAVEFORMATEX) &format;

	if (FAILED (ds_main->CreateSoundBuffer(&bufferDesc, &ds_sound_list[num_sounds], NULL))) {		
		ReportProblem (GAMEX_SOUNDCREATE_FAILED);
	}
	
	num_sounds++;
	return num_sounds-1;
}
